#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <locale.h>

#ifdef USE_WIDEC
#define _XOPEN_SOURCE_EXTENDED
#include <wctype.h>
#include <cursesw.h>
char *mb_locale = "ru_RU.KOI8-R",
     *utf_locale = "ru_RU.UTF-8";
#else
#include <curses.h>
#endif

extern "C" {
#include <libdis.h>
}

/*
  symbol 0x9b is somehow bad, isprint returns true, but
  when we output it, cursor moves to left.
*/

void set_keyhint( int n, char *hint ) {
  int hint_offs[] = { 0, 1, 9, 17, 25, 33, 41, 49, 57, 65, 74 };

  attrset( COLOR_PAIR(1) );
  mvaddstr( LINES-1, hint_offs[n], hint );
}

void disasm_init( char *fname ) {
  int i;

  init_pair( 1, COLOR_BLACK, COLOR_CYAN );
  init_pair( 2, COLOR_CYAN, COLOR_BLUE );
  init_pair( 3, COLOR_WHITE, COLOR_BLACK );
  init_pair( 4, COLOR_BLACK, COLOR_WHITE );
  init_pair( 5, COLOR_YELLOW, COLOR_BLUE );
  move( 0, 0 );
  attrset( COLOR_PAIR(1) );
  for ( i = 0; i < COLS; i++ ) addch(' ');
  attrset( COLOR_PAIR(2) );
  for ( i = 0; i < (LINES-2)*COLS; i++ ) addch( ' ' );
  attrset( COLOR_PAIR(1) );
  for ( i = 0; i < COLS; i++ ) addch(' ');
  mvaddstr( 0, 1, fname );
  addch( ' ' );
  addch( ACS_VLINE );
  attrset( COLOR_PAIR(3) );
  mvaddstr( LINES-1, 0, "1" );
  mvaddstr( LINES-1, 7, " 2" );
  mvaddstr( LINES-1, 15, " 3" );
  mvaddstr( LINES-1, 23, " 4" );
  mvaddstr( LINES-1, 31, " 5" );
  mvaddstr( LINES-1, 39, " 6" );
  mvaddstr( LINES-1, 47, " 7" );
  mvaddstr( LINES-1, 55, " 8" );
  mvaddstr( LINES-1, 63, " 9" );
  mvaddstr( LINES-1, 71, " 10" );
  attrset( COLOR_PAIR(1) );
  set_keyhint( 1, "About" );
  set_keyhint( 3, "Edit" );
  set_keyhint( 4, "Mode" );
  set_keyhint( 5, "Goto" );
  set_keyhint( 7, "Search" );
  set_keyhint( 10, "Quit" );
}

char buf[512];
int fd, buf_offs, fsize;

unsigned char *get_data( int offs ) {
  if ( offs >= fsize ) return 0;
  if ( offs < buf_offs || offs > buf_offs + sizeof(buf) - 40 ) {
  	if ( offs - (int) sizeof(buf)/2 > 0 ) buf_offs = offs - sizeof(buf)/2;
	else buf_offs = 0;
	
  	lseek( fd, buf_offs, SEEK_SET );
	read( fd, buf, sizeof buf );
  }
  return (unsigned char*)buf + offs - buf_offs;
}

int do_forw_search( int offs0, unsigned char *pat, int patlen ) {
  int i, offs = offs0;
  unsigned char *buf, *p;

  buf = (unsigned char*) malloc( patlen );
  for ( i = 0; i < patlen; i++ ) {
  	p = get_data( offs + i );
	if ( !p ) {
		free( buf );
		return -1;
	}
	buf[i] = *p;
  }

  for (;;) {
  	if ( !memcmp( pat, buf, patlen ) ) break;
	memcpy( buf, buf+1, patlen-1 );
	p = get_data( offs + patlen );
	if ( !p ) {
		free( buf );
		return -1;
	}
	buf[patlen-1] = *p;
	offs++;
  }

  free( buf );
  return offs;
}

void do_showabout() {
  WINDOW *win;
  int c;
 
  win = newwin( 6, 50, (LINES-6)/2, (COLS-50)/2 );
  wbkgd( win, COLOR_PAIR(4) );
  box( win, ACS_VLINE, ACS_HLINE );
  mvwprintw( win, 1, 1, "Disassembler v1.0" );
  mvwprintw( win, 2, 1, "(c) 2003 Copyright Vad Rulezz <vr5@mail.ru>" );
  mvwprintw( win, 3, 1, "Homepage: http://vr5.by.ru/disasm" );
  mvwprintw( win, 4, 1, "Distributed under GPL v2 or later" );
  wrefresh( win );
  move( 0, COLS-1 );
  refresh();
  for (;;) {
	c = wgetch( win );
	if ( c != ('L' & 0x1f) ) break;
	clearok( stdscr, true );
	refresh();
  }
  delwin( win );
}

void do_showhex( int offs ) {
  int i, j;
  unsigned char *p;

  offs &= ~0xf;

  attrset( COLOR_PAIR(2) | A_BOLD );
  for ( i = 0; i < LINES-2; i++ ) {
  	mvprintw( i+1, 0, "%08X  ", offs );
	for ( j = 0; j < 8; j++ ) {
		p = get_data( offs + j );
		if ( p ) printw( "%02X ", *p );
		else addstr( "   " );
	}
	addstr( "- " );
	for ( j = 0; j < 8; j++ ) {
		p = get_data( offs + 8 + j );
		if ( p ) printw( "%02X ", *p );
		else addstr( "   " );
	}
	addch( ' ' );
	for ( j = 0; j < 16; j++ ) {
		p = get_data( offs + j );
		if ( !p ) {
			addch( ' ' );
			continue;
		}
		if ( isprint( *p ) && *p != 0x9b )
			addch( *p );
		else
			addch( '.' );
	}
	offs += 16;
  }
}

void do_showtext( int offs ) {
  int i, j, rc;
  unsigned char *p;
#ifdef USE_WIDEC
  wchar_t wc[2];
  cchar_t cc;
  char mbs[MB_CUR_MAX+1];

  wc[1] = 0;
  setlocale( LC_CTYPE, mb_locale );
#endif
 
  attrset( COLOR_PAIR(2) | A_BOLD );
  move( 1, 0 );
  for ( i = 0; i < LINES-2; i++ )
  	for ( j = 0; j < COLS; j++ ) {
		p = get_data( offs );
		if ( !p ) {
			addch( ' ' );
			offs++;
			continue;
		}
#ifdef USE_WIDEC
		rc = mbtowc( &wc[0], (char*)p, 1 );
		if ( rc != -1 && iswprint( wc[0] ) )
			addwstr( wc );
#else
		if ( *p != 0x9b && isprint( *p ) )
			addch( *p );
#endif
		else addch( '.' );
		offs++;
	}
#ifdef USE_WIDEC
  setlocale( LC_CTYPE, "" );
#endif
  refresh();
}

void do_highlight_text( int offs, int patoffs, int patlen ) {
  int i;
  unsigned char *p;

  if ( patoffs == -1 ||
       patoffs + patlen < offs ||
       patoffs > offs + (LINES-2)*COLS ) return;
  if ( patoffs < offs ) {
  	patlen -= offs - patoffs;
  	patoffs = offs;
  }
  if ( patoffs + patlen >= offs + (LINES-2)*COLS )
  	patlen = offs + (LINES-2)*COLS - patoffs;
  attrset( COLOR_PAIR(3) | A_REVERSE );
  move( (patoffs-offs)/COLS+1, (patoffs-offs)%COLS );
  for ( i = 0; i < patlen; i++ ) {
  	p = get_data( patoffs + i );
	if ( !p ) {
		addch( ' ' );
		continue;
	}
  	if ( *p != 0x9b && isprint(*p) ) addch( *p ); else addch( '.' );
  }
  move( 0, COLS-1 );
}
		
void do_highlight_hex( int offs, int patoffs, int patlen ) {
  unsigned char *p;
  int i, line, pos, poffs;

  if ( patoffs == -1 ||
       patoffs + patlen < offs ||
       patoffs > offs + (LINES-2)*16 ) return;
  if ( patoffs < offs ) {
  	patlen -= offs - patoffs;
  	patoffs = offs;
  }
  if ( patoffs + patlen >= offs + (LINES-2)*16 )
  	patlen = offs + (LINES-2)*16 - patoffs;
  attrset( COLOR_PAIR(3) | A_REVERSE );
  for ( i = 0; i < patlen; i++ ) {
  	line = (patoffs + i - offs) / 16 + 1;
	pos = (patoffs + i) & 0xf;
  	p = get_data( patoffs + i );
	if ( pos > 7 ) poffs = 12; else poffs = 10;
	mvprintw( line, (pos * 3) + poffs, "%02X", *p );
	if ( i != patlen - 1 ) {
		if ( pos != 15 ) addch( ' ' );
		if ( pos == 7 ) addstr( "- " );
	}

	if ( *p != 0x9b && isprint( *p ) ) mvaddch( line, pos + 61, *p );
	else mvaddch( line, pos + 61, '.' );
  }

  move( 0, COLS-1 );
}

void do_highlight_disasm( int offs, int patoffs, int patlen ) {
  int sz, i, j;
  struct instr insn;
  unsigned char *p;

  if ( patoffs == -1 ||
       patoffs + patlen < offs ) return;
  if ( patoffs < offs ) {
  	patlen -= offs - patoffs;
  	patoffs = offs;
  }
  attrset( COLOR_PAIR(3) | A_REVERSE );
  for ( i = 0; i < LINES-2; i++ ) {
  	if ( offs > patoffs + patlen ) break;
	p = get_data( offs );
	sz = disassemble_address( (char*)p, fsize - offs, &insn );
	if ( !sz ) sz = 1;
	for ( j = 0; j < sz; j++ )
		if ( offs + j >= patoffs && offs + j < patoffs + patlen )
			mvprintw( i + 1, j * 2 + 9, "%02X", p[j] );
	offs += sz;
  }
  move( 0, COLS-1 );
}

bool quit_requested;

void edit_hexcode( int offs ) {
  bool hi_nibble, ascii_mode;
  char istr[2];
  unsigned char *buf, *p;
  int c, val, val2, coffs, x, y, i, xtbl[16];
  int asc_pos;

  x = 10;
  for ( i = 0; i < 16; i++ ) {
  	xtbl[i] = x;
	x += 3;
	if ( i == 7 ) x += 2;
  }
  asc_pos = x + 1;

  buf = (unsigned char*) malloc( 16 * LINES-2 );
  for ( i = 0; i < 16 * LINES-2; i++ ) {
  	p = get_data( offs + i );
	if ( p ) buf[i] = *p; else buf[i] = 0;
  }

  set_keyhint( 9, "Save" );

  hi_nibble = true;
  ascii_mode = false;
  coffs = 0;
  for (;;) {
  	y = coffs/16 + 1;
	if ( ascii_mode )
		x = asc_pos + coffs % 16;
	else {
		x = xtbl[coffs % 16];
		if ( !hi_nibble ) x++;
	}
	move( y, x );

get_next_key:
	c = getch();
	if ( c == KEY_RIGHT ) {
move_to_right:
		if ( ascii_mode ) {
			if ( coffs != (LINES-2) * 16 - 1 ) coffs++;
		}
		else {
			if ( !hi_nibble ) {
				if ( coffs == (LINES-2) * 16 - 1 )
					goto get_next_key;
				coffs++;
			}
			hi_nibble = !hi_nibble;
		}	
	}
	else if ( c == KEY_LEFT ) {
		if ( ascii_mode ) {
			if ( coffs > 0 ) coffs--;
		}
		else {
			if ( hi_nibble ) {
				if ( !coffs ) goto get_next_key;
				coffs--;
			}
			hi_nibble = !hi_nibble;
		}
	}
	else if ( c == KEY_DOWN ) {
		if ( coffs + 16 >= (LINES-2) * 16 ) goto get_next_key;
		coffs += 16;
	}
	else if ( c == KEY_UP ) {
		if ( coffs < 16 ) goto get_next_key;
		coffs -= 16;
	}
	else if ( c == '\t' ) {
		ascii_mode = !ascii_mode;
		if ( !ascii_mode ) hi_nibble = true;
	}
	else if ( c == KEY_BACKSPACE ) {
		if ( coffs == 0 ) goto get_next_key;

		if ( ascii_mode ) {
			coffs--;
			val = *get_data( offs + coffs );
			buf[coffs] = val;
			attrset( COLOR_PAIR(2) | A_BOLD );
			mvprintw( coffs / 16 + 1, xtbl[coffs % 16], "%02X", val );
			move( coffs / 16 + 1, asc_pos + coffs % 16 );
			if ( isprint(val) && val != 0x9b ) addch( val ); else addch( '.' );
		}
		else {
			if ( hi_nibble ) coffs--;
			hi_nibble = !hi_nibble;
			val = *get_data( offs + coffs );
			val2 = buf[coffs];
			attrset( COLOR_PAIR(2) | A_BOLD );
			if ( hi_nibble ) {
				buf[coffs] = val2 & 0xf | val & 0xf0;
				mvprintw( coffs / 16 + 1, xtbl[coffs % 16], "%X", val >> 4 );
			}
			else {
				buf[coffs] = val2 & 0xf0 | val & 0xf;
				mvprintw( coffs / 16 + 1, xtbl[coffs % 16] + 1, "%X", val & 0xf );
			}
			if ( buf[coffs] != val ) attrset( COLOR_PAIR(5) | A_BOLD );
			else attrset( COLOR_PAIR(2) | A_BOLD );
			move( coffs / 16 + 1, asc_pos + coffs % 16 );
			if ( isprint(buf[coffs]) && buf[coffs] != 0x9b )
				addch( buf[coffs] );
			else addch( '.' );
		}
	}
	else if ( c == KEY_F(9) ) {
	}
	else if ( c == KEY_F(10) ) {
		quit_requested = true;
		break;
	}
	else if ( c == ('L' & 0x1f) ) {
		clearok( stdscr, true );
		refresh();
	}
	else if ( c == 033 ) break;
	else if ( ascii_mode && c == (c & 0xff) ) {
		val2 = *get_data( offs + coffs );

		buf[coffs] = c;
		if ( val2 != c ) attrset( COLOR_PAIR(5) | A_BOLD );
		else attrset( COLOR_PAIR(2) | A_BOLD );

		if ( isprint( c ) && c != 0x9b ) addch( c );
		else addch( '.' );

		y = coffs / 16 + 1;
		x = xtbl[coffs % 16];
		move( y, x );
		if ( (val2 & 0xf0) != (c & 0xf0) )
			attrset( COLOR_PAIR(5) | A_BOLD );
		else attrset( COLOR_PAIR(2) | A_BOLD );
		printw( "%X", c >> 4 );
		if ( (val2 & 0xf) != (c & 0xf) )
			attrset( COLOR_PAIR(5) | A_BOLD );
		else attrset( COLOR_PAIR(2) | A_BOLD );
		printw( "%X", c & 0xf );

		goto move_to_right;
	}
	else if ( !ascii_mode && c == (c & 0xff) && isxdigit(c) ) {
		bool change = false;

		istr[0] = c;
		istr[1] = 0;
		val = strtol( istr, 0, 16 );
		val2 = *get_data( offs + coffs );
	
		p = &buf[coffs];
		if ( hi_nibble ) {
			val = val << 4 | (*p & 0xf);
			if ( (val & 0xf0) != (val2 & 0xf0) ) change = true;
		}
		else {
			val |= *p & 0xf0;
			if ( (val & 0xf) != (val2 & 0xf) ) change = true;
		}
		*p = val;

		if ( change ) attrset( COLOR_PAIR(5) | A_BOLD );
		else attrset( COLOR_PAIR(2) | A_BOLD );
	
		addch( toupper(c) );

		if ( val == val2 ) attrset( COLOR_PAIR(2) | A_BOLD );
		else attrset( COLOR_PAIR(5) | A_BOLD );
		move( coffs / 16 + 1, asc_pos + coffs % 16 );
		if ( isprint(buf[coffs]) && buf[coffs] != 0x9b )
			addch( buf[coffs] );
		else addch( '.' );

		goto move_to_right;
	}
  }
  free( buf );

  set_keyhint( 9, "    " );
  move( 0, 0 );
}

struct edit_disasm_state {
  int offs, line, pos, *offstbl;
  int bufsz;
  unsigned char *buf;
  bool hi_nibble;
};

void update_offstbl( edit_disasm_state& st ) {
  int i, coffs, sz;
  char *p;
  struct instr insn;
  
  coffs = st.offstbl[st.line];
  for ( i = st.line; i < LINES - 2; i++ ) {
  	st.offstbl[i] = coffs;
	if ( coffs >= st.bufsz - 20 ) {
		st.bufsz += 512;
		st.buf = (unsigned char*) realloc( st.buf, st.bufsz );
	}
  	sz = disassemble_address( (char*)&st.buf[coffs], st.bufsz - coffs, &insn );
	if ( !sz ) sz++;
	coffs += sz;
  }
  st.offstbl[i] = coffs;
}

void print_hexbyte_upd( int val, int val2 ) {
  if ( (val & 0xf0) != (val2 & 0xf0) ) {
	attrset( COLOR_PAIR(5) | A_BOLD );
	printw( "%X", val2 >> 4 );
	attrset( COLOR_PAIR(2) | A_BOLD );
  }
  else printw( "%X", val2 >> 4 );
  if ( (val & 0xf) != (val2 & 0xf) ) {
	attrset( COLOR_PAIR(5) | A_BOLD );
	printw( "%X", val2 & 0xf );
	attrset( COLOR_PAIR(2) | A_BOLD );
  }
  else printw( "%X", val2 & 0xf );
}

void update_bview( edit_disasm_state& st ) {
  int i, j, offs, val, val2, y, x, sz;
  unsigned char *p, *p2;
  struct instr insn;

  attrset( COLOR_PAIR(2) | A_BOLD );
  for ( i = st.line; i < LINES-2; i++ ) {
  	memset( &insn, 0, sizeof insn );
	offs = st.offstbl[i];
	p = &st.buf[offs];
  	sz = disassemble_address( (char*)p, st.bufsz - offs, &insn );
	mvprintw( i + 1, 0, "%08X ", st.offs + offs );
	if ( sz ) {
		for ( j = 0; j < 8; j++ )
			if ( j < sz ) {
				val = *get_data( st.offs + offs + j );
				val2 = st.buf[offs + j];
				print_hexbyte_upd( val, val2 );
			}
			else printw( "  " );
		printw( " %s", insn.mnemonic );
		if ( strlen(insn.dest) ) printw( " %s", insn.dest );
		if ( strlen(insn.src) ) printw( ", %s", insn.src );
		if ( strlen(insn.aux) ) printw( ", %s", insn.aux );
	}
	else {
		val = *get_data( st.offs + offs );
		val2 = st.buf[offs];
		print_hexbyte_upd( val, val2 );
		addstr( "               (bad)" );
	}
	getyx( stdscr, y, x );
	for ( j = x; j < COLS; j++ ) addch( ' ' );
  }
}

void edit_disasm( int offs ) {
  int c, i, val, val2, sz, coffs;
  unsigned char *p;
  char istr[2];
  bool change;
  struct instr insn;
  edit_disasm_state st;

  st.offstbl = (int*) malloc( (LINES - 1) * sizeof(int) );
  st.offs = offs;
  coffs = offs;
  for ( i = 0; i < LINES - 2; i++ ) {
  	st.offstbl[i] = coffs - offs;
  	p = get_data( coffs );
  	sz = disassemble_address( (char*)p, fsize - coffs, &insn );
	if ( !sz ) sz++;
	coffs += sz;
  }
  st.offstbl[i] = coffs - offs;
 
  st.bufsz = (coffs - offs + 511) & ~511;
  st.buf = (unsigned char*) malloc( st.bufsz );
  for ( i = 0; i < st.bufsz; i++ ) {
  	p = get_data( offs + i );
	st.buf[i] = *p;
  }

  coffs = 0;
  st.hi_nibble = true;
  st.line = 0;
  st.pos = 0;
  for (;;) {
  	move( st.line+1, st.pos+9 );
get_next_key:
	c = getch();
	if ( c == KEY_DOWN ) {
		if ( st.line+1 >= LINES-2 ) goto get_next_key;
		st.line++;
		if ( st.pos >= (st.offstbl[st.line+1] - st.offstbl[st.line]) * 2 ) {
			st.pos = (st.offstbl[st.line+1] - st.offstbl[st.line]) * 2 - 1;
			st.hi_nibble = false;
		}
	}
	else if ( c == KEY_UP ) {
		if ( !st.line ) goto get_next_key;
		st.line--;
		if ( st.pos >= (st.offstbl[st.line+1] - st.offstbl[st.line]) * 2 ) {
			st.pos = (st.offstbl[st.line+1] - st.offstbl[st.line]) * 2 - 1;
			st.hi_nibble = false;
		}
	}
	else if ( c == KEY_RIGHT ) {
move_to_right:
		if ( st.pos+1 < (st.offstbl[st.line+1] - st.offstbl[st.line]) * 2) {
			st.pos++;
			st.hi_nibble = !st.hi_nibble;
		}
		else
			if ( st.line+1 < LINES-2 ) {
				st.line++;
				st.pos = 0;
				st.hi_nibble = true;
			}
	}
	else if ( c == KEY_LEFT ) {
		if ( st.pos ) {
			st.pos--;
			st.hi_nibble = !st.hi_nibble;
		}
		else
			if ( st.line ) {
				st.line--;
				st.pos = (st.offstbl[st.line+1] - st.offstbl[st.line]) * 2 - 1;
				st.hi_nibble = false;
			}
	}
	else if ( c == KEY_BACKSPACE ) {
		if ( st.pos ) st.pos--;
		else {
			if ( st.line ) {
				st.line--;
				st.pos = (st.offstbl[st.line+1] - st.offstbl[st.line]) * 2 - 1;
			}
			else goto get_next_key;
		}
		st.hi_nibble = !st.hi_nibble;
	
		coffs = st.offstbl[st.line] + st.pos/2;
		val = st.buf[coffs];
		val2 = *get_data( offs + coffs );
		if ( st.hi_nibble )
			st.buf[coffs] = val & 0xf | val2 & 0xf0;
		else
			st.buf[coffs] = val & 0xf0 | val2 & 0xf;

		update_offstbl( st );
		update_bview( st );
	}
	else if ( c == (c & 0xff) && isxdigit(c) ) {
		istr[0] = c;
		istr[1] = 0;
		val = strtol( istr, 0, 16 );

		coffs = st.offstbl[st.line] + st.pos/2;
		val2 = *get_data( offs + coffs );
		if ( st.hi_nibble )
			st.buf[coffs] = (st.buf[coffs] & 0xf) | (val << 4);
		else
			st.buf[coffs] = (st.buf[coffs] & 0xf0) | val;

		update_offstbl( st );
		update_bview( st );
		goto move_to_right;
	}
	else if ( c == 033 ) break;
	else if ( c == KEY_F(10) ) {
		quit_requested = true;
		break;
	}
  }

  free( st.buf );
  free( st.offstbl );
}

void do_disasm( int offs ) {
  int sz, i, j, x, y;
  unsigned char *p;
  struct instr insn;

  attrset( COLOR_PAIR(2) | A_BOLD );
  for ( i = 0; i < LINES-2; i++ ) {
  	memset( &insn, 0, sizeof insn );
	p = get_data( offs );
	if ( !p ) break;
  	sz = disassemble_address( (char*)p, fsize - offs, &insn );
	if ( sz ) {
		mvprintw( i + 1, 0, "%08X ", offs );
		for ( j = 0; j < 8; j++ )
			if ( j < sz ) printw( "%02X", p[j] );
			else printw( "  " );
		printw( " %s", insn.mnemonic );
		if ( strlen(insn.dest) ) printw( " %s", insn.dest );
		if ( strlen(insn.src) ) printw( ", %s", insn.src );
		if ( strlen(insn.aux) ) printw( ", %s", insn.aux );
		offs += sz;
	}
	else {
		mvprintw( i + 1, 0, "%08X %02X               (bad)",
			offs, ((unsigned char*)p)[0] );
		offs++;
	}
	getyx( stdscr, y, x );
	for ( j = x; j < COLS; j++ ) addch( ' ' );
  }
}

void try_scroll_down( int& offs, int lines=1 ) {
  struct instr insn;
  unsigned char *p;
  int sz;

  while ( lines ) {
	p = get_data( offs );
	sz = disassemble_address( (char*)p, fsize - offs, &insn );
	if ( !sz && fsize - offs ) sz++;
	offs += sz;
	lines--;
  }
}

void try_scroll_up( int& offs, int lines=1 ) {
  int sz, lines2, offs2, offs_org = offs;
  struct instr insn;
  unsigned char *p;

  while ( offs ) {
  	offs--;
	offs2 = offs;
	lines2 = 0;
	while ( offs2 < offs_org && lines2 < lines ) {
		p = get_data( offs2 );
		sz = disassemble_address( (char*)p, fsize - offs2, &insn );
		if ( !sz ) sz++;
		offs2 += sz;
		if ( offs2 <= offs_org ) lines2++;
	}
	if ( lines2 >= lines ) break;
  }
}

void edit_asmcode(int& offs) {
}

enum disasm_mode_type {
  DISASM_MODE,
  HEX_MODE,
  TEXT_MODE
};

disasm_mode_type disasm_mode;

struct string_ent_state {
  int y, x, idx;
  bool hi_nibble;
  unsigned char *buf;
  int bsz;
  WINDOW *win;
};

void enter_offset( int& offs );
void init_xstr_entry( string_ent_state *st, WINDOW *win,
	unsigned char *buf, int bsz );
int enter_cstring( string_ent_state *st, int& c, bool single_char );
int enter_xstring( string_ent_state *st, int& c, bool single_char );
void update_cstring( string_ent_state *st );
void update_xstring( string_ent_state *st );

bool do_enter_search( unsigned char **spat, int *lpat ) {
  unsigned char *buf;
  WINDOW *win;
  int n, c, done;
  string_ent_state xst, cst;
  bool ascii = false, forward = true;
 
  if ( !spat ) return false;
 
  if ( *spat )
  	buf = *spat;
  else
  	buf = *spat = (unsigned char*) malloc( 50 );
 
  win = newwin( 6, 50, (LINES-6)/2, (COLS-50)/2 );
  keypad( win, TRUE );
  wbkgd( win, COLOR_PAIR(4) );
  box( win, ACS_VLINE, ACS_HLINE );
  mvwprintw( win, 0, 2, "Search/Forward" );
  mvwprintw( win, 1, 1, "Ascii" );
  mvwprintw( win, 3, 1, "Hex" );
  wmove( win, 2, 1 );
  init_xstr_entry( &cst, win, buf, 16 );
  cst.idx = *lpat;
  update_cstring( &cst );
  wmove( win, 4, 1 );
  init_xstr_entry( &xst, win, buf, 16 );
  xst.idx = *lpat;
  update_xstring( &xst );
  for (;;) {
  	if ( ascii ) {
		done = enter_cstring( &cst, c, true );
		xst.idx = cst.idx;
		xst.hi_nibble = true;
		update_xstring( &xst );
		if ( done ) break;
	}
	else {
		done = enter_xstring( &xst, c, true );
		cst.idx = xst.idx;
		if ( !xst.hi_nibble ) cst.idx++;
		update_cstring( &cst );
		if ( done ) break;
	}

	if ( c == KEY_UP || c == KEY_DOWN || c == '\t' )
		ascii = !ascii;
	else if ( c == KEY_F(2) ) {
		forward = !forward;
		if ( forward ) {
			mvwaddstr( win, 0, 9, "Forward" );
			waddch( win, ACS_HLINE );
		}
		else
			mvwaddstr( win, 0, 9, "Backward" );
	}
	else if ( c == ('L' & 0x1f) ) {
		clearok( stdscr, true );
		refresh();
	}
  }
  delwin( win );

  *lpat = cst.idx;
  if ( done == 1 ) return true; else return false;
}

void disasm_loop( int& offs ) {
  int c, patoffs = -1, lpat = 0;
  unsigned char *spat = 0;

  do_disasm( offs );

  while ( !quit_requested ) {
	c = getch();
	switch (c) {
	case KEY_DOWN:
		try_scroll_down( offs );
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_UP:
		try_scroll_up( offs );
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_NPAGE:
		try_scroll_down( offs, LINES-2 );
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_PPAGE:
		try_scroll_up( offs, LINES-2 );
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_HOME:
		offs = 0;
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_F(1):
		do_showabout();
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_F(3):
		edit_disasm( offs );
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_F(4):
	case '\n':
		disasm_mode = HEX_MODE;
		return;
	case KEY_F(5):
		enter_offset( offs );
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case KEY_F(17):
		if ( patoffs != -1 && patoffs < fsize-1 ) {
			patoffs = do_forw_search( patoffs+1, spat, lpat );
			goto search_done;
		}
		break;
	case KEY_F(7):
		if ( do_enter_search( &spat, &lpat ) ) {
			patoffs = do_forw_search( offs, spat, lpat );
search_done:
			if ( patoffs != -1 ) offs = patoffs;
		}
		do_disasm( offs );
		do_highlight_disasm( offs, patoffs, lpat );
		break;
	case 'L' & 0x1f:
		clearok( stdscr, true );
		refresh();
		break;
	case 033:
	case KEY_F(10):
		quit_requested = true;
		break;;
	}
  }
}

void hex_loop( int& offs ) {
  int c, lpat = 0, patoffs = -1;
  unsigned char *spat = 0;

  offs &= ~0xf;
  do_showhex( offs );

  while ( !quit_requested ) {
	c = getch();
	switch (c) {
	case KEY_DOWN:
		if ( offs + (LINES-2) * 16 < fsize ) offs += 16;
//		else {
//			if ( fsize > (LINES-2) * 16 )
//				offs = fsize - (LINES-2) * 16;
//			else offs = 0;
//		}
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_UP:
		if ( offs >= 16 ) {
			offs -= 16;
			do_showhex( offs );
			do_highlight_hex( offs, patoffs, lpat );
		}
		else offs = 0;
		break;
	case KEY_NPAGE:
		if ( offs + (LINES-2) * 16 * 2 < fsize ) offs += (LINES-2) * 16;
		else {
			if ( fsize > (LINES-2) * 16 ) {
				offs = (fsize - (LINES-2) * 16) & ~0xf;
			}
			else offs = 0;
		}

		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_PPAGE:
		if ( offs > (LINES-2) * 16 ) offs -= (LINES-2) * 16;
		else offs = 0;

		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_HOME:
		offs = 0;
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_END:
		offs = (fsize - (LINES-2) * 16 & ~0xf);
		if ( offs < 0 ) offs = 0;
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_F(1):
		do_showabout();
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_F(3):
		edit_hexcode( offs );
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_F(4):
	case '\n':
		disasm_mode = TEXT_MODE;
		return;
	case KEY_F(5):
		enter_offset( offs );
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case KEY_F(17):
		if ( patoffs != -1 && patoffs < fsize-1 ) {
			patoffs = do_forw_search( patoffs+1, spat, lpat );
			goto search_done;
		}
		break;
	case KEY_F(7):
		if ( do_enter_search( &spat, &lpat ) ) {
			patoffs = do_forw_search( offs, spat, lpat );
search_done:	
			if ( patoffs != -1 ) {
				if ( patoffs < offs ||
				     patoffs > offs + (LINES-2)*16 ) {
					offs = patoffs;
					offs -= offs & 0xf;
				}
			}
		}
		do_showhex( offs );
		do_highlight_hex( offs, patoffs, lpat );
		break;
	case 'L' & 0x1f:
		clearok( stdscr, true );
		refresh();
		break;
	case 033:
	case KEY_F(10):
		quit_requested = true;
		break;
	}
  }
}

void text_loop( int& offs ) {
  int c, patoffs = -1, lpat = 0;
  unsigned char *spat = 0;

  do_showtext( offs );

  while ( !quit_requested ) {
  	c = getch();
	switch ( c ) {
	case KEY_DOWN:
		offs += COLS;
		if ( offs + (LINES-2) * COLS > fsize ) {
			offs = fsize + COLS-1 - (LINES-2) * COLS;
			offs -= offs % COLS;
			if ( offs < 0 ) offs = 0;
		}
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case KEY_UP:
		offs -= COLS;
		if ( offs < 0 ) offs = 0;
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case KEY_NPAGE:
		offs += (LINES-2) * COLS;
		if ( offs + (LINES-2) * COLS > fsize ) {
			offs = fsize + COLS-1 - (LINES-2) * COLS;
			offs -= offs % COLS;
			if ( offs < 0 ) offs = 0;
		}
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case KEY_PPAGE:
		offs -= (LINES-2) * COLS;
		if ( offs < 0 ) offs = 0;
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case KEY_HOME:
		offs = 0;
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case KEY_END:
		offs = fsize + COLS-1 - (LINES-2) * COLS;
		offs -= offs % COLS;
		if ( offs < 0 ) offs = 0;
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case KEY_F(1):
		do_showabout();
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case KEY_F(4):
	case '\n':
		disasm_mode = DISASM_MODE;
		return;
	case KEY_F(17):
		if ( patoffs != -1 && patoffs < fsize-1 ) {
			patoffs = do_forw_search( patoffs+1, spat, lpat );
			goto search_done;
		}
		break;
	case KEY_F(7):
		if ( do_enter_search( &spat, &lpat ) ) {
			patoffs = do_forw_search( offs, spat, lpat );
search_done:
			if ( patoffs != -1 ) {
				if ( patoffs < offs ||
				     patoffs > offs + (LINES-2)*COLS ) {
					offs = patoffs;
					offs -= offs % COLS;
				}
			}
		}
		do_showtext( offs );
		do_highlight_text( offs, patoffs, lpat );
		break;
	case 'L' & 0x1f:
		clearok( stdscr, true );
		refresh();
		break;
	case 033:
	case KEY_F(10):
		quit_requested = true;
		break;
	}
  }
}

void enter_offset( int& offs ) {
  int i, c, req_offs;
  char buf[9], *p;

  move( 1, 0 );
  for ( i = 0; i < 8; i++ ) addch( ACS_CKBOARD );

  move ( 1, 0 );
  p = buf;
  for (;;) {
  	c = getch();
	if ( c == (c & 0xff) && isxdigit(c) && p < buf + 8 ) {
		c = toupper(c);
		addch( c );
		*p++ = c;
	}
	else if ( c == KEY_BACKSPACE ) {
		p--;
		mvaddch( 1, p - buf, ACS_CKBOARD );
		move( 1, p - buf );
	}
	else if ( c == '\n' ) {
		*p = 0;
		break;
	}
	else if ( c == ('L' & 0x1f) ) {
		clearok( stdscr, true );
		refresh();
	}
	else if ( c == 033 )
		return;
  }

  req_offs = strtol( buf, 0, 16 );
  if ( req_offs >= 0 && req_offs < fsize ) offs = req_offs;
}

void init_xstr_entry( string_ent_state *st, WINDOW *win,
	unsigned char *buf, int bsz ) {
  st->win = win;
  st->buf = buf;
  st->bsz = bsz;
  
  getyx( st->win, st->y, st->x );
  whline( st->win, ACS_CKBOARD, st->bsz * 3 - 1 );
  wmove( st->win, st->y, st->x );
  st->idx = 0;
  st->hi_nibble = true;
}

int enter_cstring( string_ent_state *st, int& c, bool single_char ) {
  wmove( st->win, st->y, st->x + st->idx );
  for (;;) {
  	c = wgetch( st->win );
	if ( c == '\n' ) return 1;
	else if ( c == 033 ) return 2;
	else if ( c == '\t' ) return 0;
	else if ( c == (c & 0xff) && st->idx < st->bsz ) {
		if ( c != 0x9b && isprint( c ) ) waddch( st->win, c );
		else waddch( st->win, '.' );
		st->buf[st->idx++] = c;
	}
	else if ( c == KEY_BACKSPACE && st->idx ) {
		st->idx--;
		mvwaddch( st->win, st->y, st->x + st->idx, ACS_CKBOARD );
		wmove( st->win, st->y, st->x + st->idx );
	}
	else break;
	if ( single_char ) break;
  }

  return 0;
}

void update_cstring( string_ent_state *st ) {
  int i, c;

  wmove( st->win, st->y, st->x );
  for ( i = 0; i < st->idx; i++ ) {
  	c = st->buf[i];
	if ( c == 0x9b || !isprint( c ) ) c = '.';
	waddch( st->win, c );
  }
  for (; i < st->bsz; i++ ) waddch( st->win, ACS_CKBOARD );
  wrefresh( st->win );
}

int enter_xstring( string_ent_state *st, int& c, bool single_char ) {
  int n, xpos;
  char s[2];

  if ( st->hi_nibble ) {
  	if ( st->idx ) xpos = st->x + st->idx * 3 - 1;
	else xpos = st->x;
  }
  else xpos = st->x + st->idx * 3 + 1;
  wmove( st->win, st->y, xpos );

  for (;;) {
  	c = wgetch( st->win );
	if ( c == (c & 0xff) && isxdigit(c) && st->idx < st->bsz ) {
		s[0] = c;
		s[1] = 0;
		n = strtol( s, 0, 16 );
		if ( st->hi_nibble ) {
			if ( st->idx ) waddch( st->win, ' ' );
			waddch( st->win, toupper(c) );
			st->buf[st->idx] = n << 4;
		}
		else {
			waddch( st->win, toupper(c) );
			n |= st->buf[st->idx] & 0xf0;
			st->buf[st->idx++] = n;
		}
		st->hi_nibble = !st->hi_nibble;
	}
	else if ( c == KEY_BACKSPACE ) {
		if ( st->hi_nibble ) {
			if ( !st->idx ) continue;
			mvwaddch( st->win, st->y, st->x + st->idx * 3 - 2,
				ACS_CKBOARD );
			wmove( st->win, st->y, st->x + st->idx * 3 - 2 );
			st->idx--;
			st->buf[st->idx] &= 0xf0;
		}
		else {
			if ( st->idx ) {
				mvwaddch( st->win, st->y,
					st->x + st->idx * 3 - 1, ACS_CKBOARD );
				waddch( st->win, ACS_CKBOARD );
				wmove( st->win, st->y,
					st->x + st->idx * 3 - 1 );
			}
			else {
				mvwaddch( st->win, st->y, st->x, ACS_CKBOARD );
				wmove( st->win, st->y, st->x );
			}
		}
		st->hi_nibble = !st->hi_nibble;
	}
	else if ( st->hi_nibble && c == '\n' ) return 1;
	else if ( c == 033 ) return 2;
	else break;
	if ( single_char ) break;
  }
  return 0;
}

void update_xstring( string_ent_state *st ) {
  int i, c;

  wmove( st->win, st->y, st->x );
  for ( i = 0; i < st->idx; i++ ) {
  	c = st->buf[i];
	if ( i ) waddch( st->win, ' ' );
	wprintw( st->win, "%02X", c );
  }
  for (; i < st->bsz; i++ ) {
  	waddch( st->win, ACS_CKBOARD );
  	waddch( st->win, ACS_CKBOARD );
  	if ( i != st->bsz-1 ) waddch( st->win, ACS_CKBOARD );
  }
  wrefresh( st->win );
}

int main( int argc, char **argv ) {
  int offs;

  if ( argc < 2 ) {
  	puts( "Syntax: ./disass <filename>" );
	return 1;
  }
  fd = open( argv[1], O_RDONLY );
  if ( fd == -1 ) {
  	perror( "disassm open" );
	return 1;
  }
  read( fd, buf, sizeof buf );
  fsize = lseek( fd, 0, SEEK_END );
  buf_offs = 0;

  setlocale( LC_CTYPE, "" );

  initscr();
  start_color();
  cbreak();
  noecho();
  disasm_init( argv[1] );
  keypad( stdscr, true );

  quit_requested = false;
  disasm_mode = DISASM_MODE;
  disassemble_init( 0, INTEL_SYNTAX );
  offs = 0;

  while ( !quit_requested ) {
  	if ( disasm_mode == DISASM_MODE )
		disasm_loop( offs );
	else if ( disasm_mode == HEX_MODE )
		hex_loop( offs );
	else
		text_loop( offs );
  }

do_exit: 
  erase();
  refresh();
  endwin();
}
